//	Joystick.h - Class to handle DirectInput Joysticks (Windows only).
//				 This code's a bit messy, I'm afraid...
//	--------------------------------------------------------------------------
//	This file is part of 'Niall's Pedal Board', a simple ASIO VST host.
//	Copyright (C) 2004-2005  Niall Moody
//	
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//	
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//	
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//	
//	--------------------------------------------------------------------------
//	
//	You can e-mail me at niallmoody@yahoo.co.uk if you want.
//	
//	--------------------------------------------------------------------------

#ifndef JOYSTICK_H_
#define JOYSTICK_H_

#include "dinput.h"

//It is VITAL that you include the JUCE headers after the windows (dinput.h)
//ones - otherwise you get nasty compilation errors.  This applies to all the
//files that reference this one (i.e. always include Joystick.h before you
//include any headers that include ../juce.h).
#include <juce.h>

#include "ParameterFilter.h"

///	Enumeration representing the various types of Joystick event possible.
typedef enum
{
	evAxis,
	evRotation,
	evSlider,
	evPOV,
	evButton,
	evVelocity,
	evAngularVelocity,
	evSliderVelocity,
	evAcceleration,
	evAngularAcceleration,
	evSliderAcceleration,
	evForce,
	evTorque,
	evSliderForce,
	evEnvelopeDetector,	//Not actually a joystick event, but it's easier if we pretend it is.
	evMIDICC,			//ditto...
	evMIDINoteVal,
	evMIDIVelocity
} JoystickEventType;

///	Class representing a joystick event.
/*!
	Includes methods to determine what kind of event it is (button, axis,
	etc.), and what the various values have been set to for this event.
 */
class JoystickEvent
{
  public:
	///	Default Constructor.
	/*!
		You'd typically call create() after you'd created the event using
		this constructor, to set the appropriate variables.
	 */
	JoystickEvent() {};
	///	Constructor.
	/*!
		\param type What type of event it is (e.g. button, axis, etc.)
		\param index e.g. which button or axis has changed.
		\param val The value of the joystick control which has changed.

		Note that any control values other than the indexed one will be set to
		0 in this method (i.e. they won't represent the joystick's actual
		state).
	 */
	JoystickEvent(JoystickEventType type, int index, int val);

	///	For 2-stage construction, together with the default constructor.
	/*!
		\param type What type of event it is (e.g. button, axis, etc.)
		\param index e.g. which button or axis has changed.
		\param val The value of the joystick control which has changed.

		Note that any control values other than the indexed one will be set to
		0 in this method (i.e. they won't represent the joystick's actual
		state).
	 */
	void create(JoystickEventType type, int index, int val);

	///	Returns the type of this event (button, axis, etc.).
	JoystickEventType getType() {return Type;};
	///	Index is e.g. which button out of the 128 has changed.
	int getIndex() {return Index;};

	///	The value of the joystick control which has changed to generate this event.
	int getVal();

	///	Returns the value for the num axis.
	int getAxis(int num) {return Axis[num];};
	///	Returns the value for the num rotation.
	int getRotation(int num) {return Rotation[num];};
	///	Returns the value for the num slider.
	int getSlider(int num) {return Slider[num];};
	///	Returns the value for the num Point of View hat.
	int getPOV(int num) {return POV[num];};
	///	Returns the value for the num button.
	bool getButton(int num) {return Button[num];};
	///	Returns the value for the num velocity.
	int getVelocity(int num) {return Velocity[num];};
	///	Returns the value for the num angular velocity.
	int getAngularVelocity(int num) {return AngularVelocity[num];};
	///	Returns the value for the num slider velocity.
	int getSliderVelocity(int num) {return SliderVelocity[num];};
	///	Returns the value for the num acceleration.
	int getAcceleration(int num) {return Acceleration[num];};
	///	Returns the value for the num angular acceleration.
	int getAngularAcceleration(int num) {return AngularAcceleration[num];};
	///	Returns the value for the num slider acceleration.
	int getSliderAcceleration(int num) {return SliderAcceleration[num];};
	///	Returns the value for the num force.
	int getForce(int num) {return Force[num];};
	///	Returns the value for the num torque.
	int getTorque(int num) {return Torque[num];};
	///	Returns the value for the num slider force.
	int getSliderForce(int num) {return SliderForce[num];};
  private:
	JoystickEventType Type;
	int Index;

	int Axis[3];
	int Rotation[3];
	int Slider[2];
	int POV[4];
	bool Button[128];
	int Velocity[3];
	int AngularVelocity[3];
	int SliderVelocity[2];
	int Acceleration[3];
	int AngularAcceleration[3];
	int SliderAcceleration[2];
	int Force[3];
	int Torque[3];
	int SliderForce[2];
};

///	Class representing a joystick.
/*!
	To use this, you'd inherit it in a Component which you're adding to the
	desktop (this is necessary - if the Component's not added to the desktop,
	this class won't work.

	Once the component's been added to the desktop, you have to call the
	enumerateDevices() method.

	After that, all you have to do is implement your own version of the
	joystickCallback() method, to handle the joystick events you get.

	Note that the way this class works is that it continuously polls the
	joystick to determine it's state, so you should expect a lot of calls
	to joystickCallback() where nothing's happened.

	\todo Make sure that joystickCallback is only called when something's
	actually changed.
 */
class Joystick : public Timer
{
  public:
	///	Default Constructor.
	/*!
		This is so you can have a Component which also inherits from Joystick,
		to make use of it's joystickCallback() method.
		The Component has to be added to the desktop, after which you have to
		call the enumerateDevices() method.
	 */
	Joystick();

	///	Constructor.
	/*!
		You don't have to call the enumerateDevices() method for this one,
		though the parent component has to be added to the desktop before
		this constructor is called.  This one also only chooses the first
		joystick it finds.
	 */
	Joystick(Component *parent, TextEditor *debug=0);
	///	Destructor.
	virtual ~Joystick();

	//Thread stuff
	//void run();

	///	Fills JoystickNameList with the names of all the joysticks attached to the system, and returns them.
	/*!
		Also handles a lot of the setup stuff, by calling the
		DIEnumDevCallback callback function.
	 */
	StringArray enumerateDevices();
	///	Sets up various joystick settings, starts polling the joystick for events.
	void acquireJoystick(String name);

	///	Returns the value of the EnumerateOnly variable.
	/*!
		EnumerateOnly distinguishes between the multiple inheritance case
		(default constructor), where the enumerateDevices() method is purely
		there to enumerate the joysticks attached to the system, and the
		other case, where the DIEnumDevCallback function also sets up
		various things which are set elsewhere in the multiple inheritance
		case.
	 */
	bool getEnumerateOnly() {return EnumerateOnly;};
	///	Adds pinst to the JoystickList and JoystickNameList arrays.
	void addToJoystickList(String joyname, LPCDIDEVICEINSTANCE pinst);

	///	Used for debugging purposes, to stream text to a TextEditor Component.
	void addText(String& text);
	///	Creates a joystick device according to our specifications.
	/*!
		This is called from the DIEnumDevCallback function, but not for the
		multiple inheritance case.
	 */
	void createJoystickDevice(LPCDIDEVICEINSTANCE pinst);
	///	Acquires a joystick (held in DirectInputDevice) for us.
	/*!
		This is called from the DIEnumDevCallback function, but not for the
		multiple inheritance case.
	 */
	void acquireJoystick();

	///	Callback method.
	/*!
		This is where you'd put you're joystick event handling code if you're
		inheriting from Joystick.

		\param index Because one Joystick object can handle more than one actual
		joystick.
	 */
	virtual void joystickCallback(JoystickEvent& joystate, int index);

	///	Sets whether or not a filter should be applied to the value of joystick events.
	/*!
		Useful if you're using fairly rubbish pots, like me.
	 */
	void setFilter(bool val) {FilterOn = val;};

	///	Sets the Component we use to get the HWND we need.
	void setParentThing(Component *parent) {Parent = parent;};
  protected:
	///	This is called to poll the joystick for any changes in state.
	void timerCallback();
	///	Creates a JoystickEvent from the current state of the joystick.
	void setEvent(DIJOYSTATE2 state, OwnedArray<JoystickEvent>& events);
	///	Sets the initial state of the joystick.
	/*!
		(sets everything to 0, basically)
	 */
	void initState();

	///	Parent *must* be a desktop window!!!
	Component *Parent;
	///	For debug purposes - the plugin will write a bunch of data to this Component.
	TextEditor *Ed;

	LPDIRECTINPUT DirectInput;
	Array<LPDIRECTINPUTDEVICE> DirectInputDevice;

	///	This is used by setEvent() to determine whether the joystick's state has changed.
	DIJOYSTATE2 OldState;

	StringArray JoystickNameList;
	Array<LPCDIDEVICEINSTANCE> JoystickList;
	bool EnumerateOnly;

	///	Used to smooth the values of certain joystick events.
	/*!
		One of these will be constructed for every control on the joystick.
	 */
	ParameterFilter **JoyFilter;
	///	Whether or not the ParameterFilter should be used or not.
	bool FilterOn;
	///	This is the current value if the filter's on.
	DIJOYSTATE2 FiltState;

	enum
	{
		AxisX,
		AxisY,
		AxisZ,
		RotationX,
		RotationY,
		RotationZ,
		SliderU,
		SliderV,
		VelocityX,
		VelocityY,
		VelocityZ,
		AngularVelocityX,
		AngularVelocityY,
		AngularVelocityZ,
		SliderVelocityU,
		SliderVelocityV,
		AccelerationX,
		AccelerationY,
		AccelerationZ,
		AngularAccelerationX,
		AngularAccelerationY,
		AngularAccelerationZ,
		SliderAccelerationU,
		SliderAccelerationV,
		ForceX,
		ForceY,
		ForceZ,
		TorqueX,
		TorqueY,
		TorqueZ,
		SliderForceU,
		SliderForceV,
		NumControls,
	};
};

#endif
