//	NPBPluginOrganiser.h - Keeps track of all the plugins loaded, and
//						   handles all the audio & control stuff too.
//	--------------------------------------------------------------------------
//	This file is part of 'Niall's Pedal Board', a simple ASIO VST host.
//	Copyright (C) 2004-2005  Niall Moody
//	
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//	
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//	
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//	
//	--------------------------------------------------------------------------
//	
//	You can e-mail me at niallmoody@yahoo.co.uk if you want.
//	
//	--------------------------------------------------------------------------

#ifndef NPBPLUGINORGANISER_H_
#define NPBPLUGINORGANISER_H_

class NPBPluginOrganiser;

#ifndef NPBWINDOW_H_
#include "NPBWindow.h"
#endif

#include <juce.h>

//#include "NPBPlugin.h"

#include "NPBPluginFactory.h"

struct AEffect;

///	The core of the program.
/*!
	This keeps an array of all plugins currently running, passes joystick
	events to them, sets up the soundcard, and handles the loading and saving
	of plugin configurations, as well as the option of recording the output
	as a .wav file.
 */
class NPBPluginOrganiser : public Component,
						   public ActionListener,
						   public AudioIODeviceCallback,
						   public MidiInputCallback
{
  public:
	///	Constructor.
	NPBPluginOrganiser();
	///	Destructor.
	~NPBPluginOrganiser();

	///	Fills in the component with a GradientBrush, based on backColour.
	void paint(Graphics& g);
	///	Not currently used.
	void actionListenerCallback(const String& m);

	///	The soundcard callback.
	/*!
		This method basically calls the ProcessAudio() methods of each
		instantiated NPBPluginBase in turn, and also handles writing the
		output to a .wav file, if the Organiser's recording.

		\sa NPBPluginBase.
	 */
	void audioDeviceIOCallback(const float **inputChannelData, int totalNumInputChannels, float **outputChannelData, int totalNumOutputChannels, int numSamples);

	///	The joystick callback.
	/*!
		Basically passes on all joystick events to all the instantiated
		plugins, in turn.
	 */
	void joystickCallback(JoystickEvent& joystate, int index);

	/// The MIDI callback.
	/*!
		Passes any MIDI input messages onto all the plugins in turn.

		\todo Allow plugins to specify they don't want to receive MIDI events
		from the MIDI input (?).
	 */
	void handleIncomingMidiMessage(MidiInput *source, const MidiMessage &message);

	///	Used to determine when to display the popup menu used to select plugins.
	void mouseDown(const MouseEvent& e);

	///	Used to determine when to delete a plugin.
	/*!
		Plugins send the Organiser a 'DELETE_THIS_PLUGIN' command message when
		they're supposed to be deleted (because they are deleted when the user
		right clicks on them and selects 'Delete this Plugin'.
	 */
	void handleCommandMessage(int commandId);

	///	So we can drag & drop dlls to add plugins.
	bool filesDropped(const StringArray& filenames, int mouseX, int mouseY);

	///	Sets the soundcard configuration.
	/*!
		This is called from NPBWindow when the soundcard configuration has
		changed.
	 */
	void setIOs(String soundcard, BitArray input, BitArray output);

	///	Not used yet - I'm not sure how to handle this nicely...
	void vstIdle();

	///	This one's for restarting audio after loading a plugin...
	void StartAudio();
	///	This one's for when the 'Start Audio' button's clicked...
	bool StartAudio(String soundcard, BitArray inputs, BitArray outputs);
	///	Stops the audio playing.
	void StopAudio();
	///	Returns the current audio block size used in the audioDeviceIOCallback() method.
	/*!
		We use this in NPBPlugin to inform the VST plugin of the current block
		size.
	 */
	int getBlockSize();
	///	Returns whether or not we're using direct mode for the audio or not.
	bool getDirectMode() {return directMode;};
	///	Sets whether or not we're using direct mode for the audio.
	void setDirectMode(bool val);

	///	Returns the names of all the currently-connected joysticks.
	/*!
		(where is this used?)
	 */
	StringArray getJoysticks();

	///	Called when the user selects New from the window's file menu
	void clearOrganiser();

	///	Saves the current configuration of plugins as a .npb xml file.
	void savePatch();
	///	Loads a previously-saved configuration of plugins (*.npb).
	void loadPatch();

	///	Tells the Organiser to start recording the audio, saving it to file.
	/*!
		Called from NPBWindow.
	 */
	bool startRecording(File &file);
	///	Tells the Organiser to stop recording the audio.
	/*!
		Called from NPBWindow.
	 */
	void stopRecording();

	///	Called from the ComponentFactory when the Organiser's instantiated.
	void setBackColour(Colour col);
	///	Called from the ComponentFactory when the Organiser's instantiated.
	void setPlugColour(Colour col);
	///	Called from the ComponentFactory when the Organiser's instantiated.
	void setPlugButtonColour(Colour col);

	///	Called from plugins to find out what colour their background should be.
	inline Colour getPlugColour() {return pluginColour;};
	///	Called from plugins to find out what colour their buttons should be.
	inline Colour getPlugButtonColour() {return pluginButtonColour;};

	///	Called from the vst callback, to tell the associated NPBPlugin the plugin requires MIDI events.
	NPBPluginBase *findPlugin(AEffect *plugin);

	///	Sets the envelope decrement for the plugins' envelope detectors.
	void setEnvelopeDecrement(float val);
	///	Returns the envelope decrement for the plugins' envelope detectors.
	float getEnvelopeDecrement() {return envelopeDecrement;};

	///	Sets the BPM we pass onto interested plugins.
	void setBpm(float val);
	///	Returns the BPM we're using.
	float getBpm() {return bpm;};

	///	Sets midiDevice according to the user's choice of input.
	void setMIDIInputDevice(int index);
	///	Returns the value of the midiDevice we're currently using.
	int getMIDIInputDevice() {return midiDeviceIndex;};
  private:
	///	This Fills Factory with all the NPBPluginBase subclasses which the program currently has.
	/*!
		The aim behind the use of a factory was to allow the program to 
		potentially be extended through the creation of different
		NPBPluginBase subclasses.

		The point of this approach is that you only have to update a single
		method of *this* class to add new NPBPluginBase classes, rather than
		having to go through the code and update menu entries etc.

		As it is, all the methods which need to know about the different kinds
		of NPBPluginBase subclasses available simply interrogate Factory to
		get the necessary information, instead of it being hard-coded into a
		number of different methods.
	 */
	void SetUpFactory();

	///	An array of all the currently-instantiated NPBPluginBases.
	OwnedArray<NPBPluginBase> PluginArray;
	///	Contains an NPBPluginBaseFactory subclass for each available NPBPluginBase subclass.
	/*!
		Every NPBPluginBase subclass which wants to make itself known to
		the Organiser has to have a static Factory object.

		Basically, Factory is a template class which handles the creation
		of objects derived from NPBPluginBase.  It is designed so that all you
		need to do to allow the Factory to construct an object is provide a
		static Factory object in the class declaration, like so:
		\code
		static Factory<Base, Child> myFactory;
		\endcode
		Where Base would be NPBPluginBase in this case, and Child would be the
		name of the subclass the factory resides in.

		So basically, Factory is just an array of these factories, and to
		create a new instance of an NPBPluginBase subclass, we just call:
		\code
		Factory[index]->createInstance(this);
		\endcode

		\sa FactoryBase, NPBPlugin
	 */
	Array<NPBPluginBaseFactory *> Factory;

	///	Represents the currently-used sound device.
	ASIOAudioIODevice *audio;

	///	The name of the current soundcard.
	String Soundcard;
	///	Specifies which inputs of the soundcard are being used.
	BitArray Inputs;
	///	Specifies which outputs of the soundcard are being used.
	BitArray Outputs;
	///	Specifies how many channels of the soundcard are being used.
	int NumChannels;
	///	Specifies the current buffer size for the soundcard.
	int BufferSize;
	///	Specifies whether the ASIO driver should be using 'direct mode' or not.
	bool directMode;

	///	The current input buffer.
	/*!
		Essentially an array (number of input channels) of an array (number of
		samples in the current buffer size).
	 */
	float **InBuffer;
	///	The current output buffer.
	/*!
		Essentially an array (number of output channels) of an array (number
		of samples in the current buffer size).
	 */
	float **OutBuffer;

	///	Lets us know we're currently recording.
	bool isRecording;
	///	Used to write the output to a .wav file when we're recording it.
	AudioFormatWriter *recordWriter;
	///	An array to hold the audio data we're writing to the .wav file, when we record it.
	/*!	The buffer we copy the float data into from the current process block
		each time, to write to the .wav file.

		(do it this way so we don't have to keep allocating memory in the
		process method)
	 */
	int **recordArray;

	///	The main colour used to colour in the background.
	Colour backColour;
	///	The main colour of the plugins displayed in the organiser.
	Colour pluginColour;
	///	The main colour of the buttons used for the plugins.
	Colour pluginButtonColour;

	///	The envelope decrement for the plugins' envelope detectors.
	float envelopeDecrement;

	///	The BPM we pass onto interested plugins.
	float bpm;

	///	The index of the MIDI input device we're currently using.
	int midiDeviceIndex;
	///	The MIDI input device we're currently using.
	MidiInput *midiDevice;
};

#endif
